#
# This file is part of "Cut the Crap" (CTC) filtering proxy server.
# Copyright 1999 Constantinos A. Kotsokalis <ckotso@cs.ntua.gr>
#
# CTC is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# CTC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CTC; see the file COPYING. If not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#

import socket, select, string, urlparse, time, sys, re, sockpair

# Called on errors where no message is sent to clients (e.g. send's
# and recv's that cause exceptions
def PrintErr(str, msg):
	try:
		errstr = msg[1]
	except:
		errstr = msg[0]
	sys.stderr.write(str + errstr + "\n")


# Function to return network error messages to the browser
def SocketErrorMsg(cfg, msg, rem_srv, rem_port):

	err_msg = cfg['ERR_MSG']
	err_msg = string.replace(err_msg, '%SERVER', rem_srv)
	err_msg = string.replace(err_msg, '%PORT', `rem_port`)
	try:
		errstr = msg[1]
	except:
		try: 
			errstr = msg[0]
		except:
			errstr = 'Unknown error'
	err_msg = string.replace(err_msg, '%ERROR', errstr)
	err_msg = string.replace(err_msg, '%DATE', time.ctime(time.time()))
	err_msg_size = len(err_msg)
	return_buff = """HTTP/1.1 403 ERROR
Server: Cut The Crap\nContent-Length: """ + `err_msg_size` + """
Content-Type: text/html
Proxy-Connection: Close

"""
	return_buff = return_buff+err_msg

	return return_buff


# Initialise listening socket with a queue of 32 connections
def init_socket(cfg):

	inlist = []

	# socket.IPPROTO_IP is not defined in python versions earlier than 1.5.2
	if 'IPPROTO_IP' in dir(socket):
		IPPROTO_IP = socket.IPPROTO_IP
	else:
		IPPROTO_IP = 0

	try:
		ls = socket.socket(socket.AF_INET, socket.SOCK_STREAM, IPPROTO_IP)
		# Ignore TIME_WAIT socket states to restart at once, when needed
		ls.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		addr = '', cfg['BIND_PORT']
		ls.bind(addr)
		ls.listen(32)
	except socket.error, msg:
		sys.stderr.write(msg[1] + "\n")
		sys.exit(msg[0])

	inlist.append(ls)

	return cfg, inlist


# Main program loop: Accept incoming connections and handle requests
def loop(cfg, inlist):

	pairs = {}
	zeng = cfg['Z_ENGINE']
	conntype = re.compile("connection: keep-alive", re.I)
	gifloc = re.compile(".*/CTC-images/CTC.gif")

	# ``MOVE'' is used by mod_roaming, let's include that one too.
	# The other ones are HTTP/1.1 compatible.
	HTTP_METHODS = ("GET", "POST", "PUT", "HEAD", "MOVE", \
					"OPTIONS", "DELETE", "TRACE", "CONNECT")

	while 1:
		try:
			rin = select.select(inlist, [], [])[0]
		except select.error, msg:
			PrintErr("select: ", msg)
			sys.exit(0)

		for sock in rin:
			if sock == inlist[0]:  # listening socket
				try:
					newconn = sock.accept()[0]
					inlist.append(newconn)
					pairs[newconn] = sockpair.SockPair(sockpair.CLIENT)
				except socket.error, msg:
					PrintErr("accept: ", msg)
				continue

			try:
				buff = sock.recv(cfg['BUFF_SIZE'])
			except socket.error, msg:
				PrintErr("recv: ", msg)
				continue

			if not len(buff):
				other_end = pairs[sock].pair
				if other_end == None:
					pass
				else:
					inlist.remove(other_end)
					del(pairs[other_end])
					other_end.close()
				inlist.remove(sock)
				del(pairs[sock])
				sock.close()
				continue

			if pairs[sock].type == sockpair.SERVER:
				other_end = pairs[sock].pair
				if other_end == None:
					inlist.remove(sock)
					del(pairs[sock])
					sock.close()
				else:
					if conntype.search(buff) != None:
						buff = conntype.sub("Proxy-Connection: Close", buff)
					try:
						other_end.send(buff)
					except socket.error, msg:
						PrintErr("send: ", msg)
			else:
				tokens = string.split(buff)
				# Will have to think about this: POST operations might
				# take place in two steps.
				if len(tokens) < 1:
					continue

				if tokens[0] in HTTP_METHODS:
			 		if zeng.urlblocker(tokens[1]):
						# send blank gif
						try:
							sock.send(cfg['ADV_RET_BUFF'])
							sock.send(cfg['RETGIF'])
						except socket.error, msg:
							PrintErr("send: ", msg)
						continue

					if gifloc.match(tokens[1]):
						try:
							sock.send(cfg['CTCGIF_RET_BUFF'])
							sock.send(cfg['CTCGIF'])
						except socket.error, msg:
							PrintErr("send: ", msg)
						continue

					if not len(cfg['PARENT_PROXY_HOST']):
						url_tuple = urlparse.urlparse(tokens[1])
						host_port = url_tuple[1]
						if (string.find(host_port, ':') != -1):
							occ = string.count(host_port, ':')
							atocc = string.count(host_port, '@')
							if occ == 1 and atocc == 0:
								rem_srv, rem_port = string.split(host_port, ':')
								rem_port = string.atoi(rem_port)
							elif occ == 1 and atocc == 1:
								rem_srv, rem_port = string.split(host_port, '@')[1], 80
							elif occ == 2:
								rem_srv, rem_port = string.split(string.split(host_port, '@')[1], ':')
								rem_port = string.atoi(rem_port)
						else:
							rem_srv, rem_port = host_port, 80

						URL = url_tuple[2]
						if len(url_tuple[3]):
							URL = URL + ';' + url_tuple[3]
						if len(url_tuple[4]):
							URL = URL + '?' + url_tuple[4]
						buff = string.replace(buff, tokens[1], URL)

					else:
						rem_srv = cfg['PARENT_PROXY_HOST']
						rem_port = cfg['PARENT_PROXY_PORT']

					try:
						server = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
					except socket.error, msg:
						PrintErr("socket: ", msg)
						inlist.remove(sock)
						del(pairs[sock])
						sock.close()
						continue

					try:
						server.connect(rem_srv, rem_port)
						pairs[server] = sockpair.SockPair(sockpair.SERVER)
						pairs[server].pair = sock
						pairs[sock].pair = server
						inlist.append(server)
					except socket.error, msg:
						PrintErr("connect: ", msg)
						try:
							sock.send(SocketErrorMsg(cfg, msg, rem_srv, rem_port))
						except socket.error, msg:
							PrintErr("send: ", msg)
						continue

					try:
						server.send(buff)
					except socket.error, msg:
						PrintErr("send: ", msg)
						try:
							sock.send(SocketErrorMsg(cfg, msg, rem_srv, rem_port))
						except socket.error, msg:
							PrintErr("send: ", msg)
						continue
				
				else: # More data from POST or PUT operation
					other_end = pairs[sock].pair
					if other_end == None: 
						# Server closed connection before getting all data
						inlist.remove(sock)
						del(pairs[sock])
						sock.close()
					else:
						try:
							other_end.send(buff)
						except socket.error, msg:
							PrintErr("send: ", msg)
					continue


